#!/usr/bin/env python
# -*- coding:utf-8 -*-


##############################################################################
## license :
##============================================================================
##
## File :        Smaract_mcs2.py
##
## Project :    SmaractMCS2
##
## This file is part of Tango device class.
##
## Tango is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Tango is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Tango.  If not, see <http://www.gnu.org/licenses/>.
##
##
## $Author :      SPE - www.elettra.eu
##
## $Revision :    $
##
## $Date :        2019
##
## $HeadUrl :     $
##
##############################################################################

__all__ = ["Smaract_mcs2", "Smaract_mcs2Class", "main"]

__docformat__ = 'restructuredtext'

import PyTango
import sys
# Add additional import
import time
import math
from distutils.util import strtobool

# SENSOR_TYPE_S = 300      # linear positioner with nano sensor (1 @ MCS1)
# SENSOR_TYPE_SR = 312     # rotary positioner with nano sensor (2 @ MCS1)
# SENSOR_TYPE_SR20 = 309   # rotary positioner with nano sensor (8 @ MCS1)
# SENSOR_TYPE_SD = 301     # extended scanning Range linear positioner with nano sensor (21 @ MCS1)
# SENSOR_TYPE_SR2 = 313    # rotary positioner with nano sensor for high applied masses (23 @ MCS1)
# SENSOR_TYPE_G605DS = 325 # goniometer with nano sensor (60.5mm radius), large actuator (48 @ MCS1)
# SENSOR_TYPE_G775DS = 328 # goniometer with nano sensor (77.5mm radius), large actuator (49 @ MCS1)
CONTROLLER_FLAGS = [
    "ACTIVELY_MOVING",          # 0x00000001
    "CLOSED_LOOP_ACTIVE",       # 0x00000002
    "CALIBRATING",              # 0x00000004
    "REFERENCING",              # 0x00000008
    "MOVE_DELAYED",             # 0x00000010
    "SENSOR_PRESENT",           # 0x00000020
    "IS_CALIBRATED",            # 0x00000040
    "IS_REFERENCED",            # 0x00000080
    "END_STOP_REACHED",         # 0x00000100
    "RANGE_LIMIT_REACHED",      # 0x00000200
    "FOLLOWING_LIMIT_REACHED",  # 0x00000400
    "MOVEMENT_FAILED",          # 0x00000800
    "IS_STREAMING",             # 0x00001000
    "POSITIONER_OVERLOAD",      # 0x00002000
    "OVER_TEMPERATURE",         # 0x00004000
    "REFERENCE_MARK",           # 0x00008000
    "IS_PHASED",                # 0x00010000
    "POSITIONER_FAULT",         # 0x00020000
    "AMPLIFIER_ENABLED",        # 0x00040000
    "UNDOCUMENTED",             # 0x00080000
    "UNDOCUMENTED",             # 0x00100000
    "UNDOCUMENTED",             # 0x00200000
    "UNDOCUMENTED",             # 0x00400000
    "UNDOCUMENTED",             # 0x00800000
    "UNDOCUMENTED"              # 0x01000000
]
END_OF_LINE = '\015\012'

## Device States Description

class Smaract_mcs2 (PyTango.Device_4Impl):

    #--------- Add you global variables here --------------------------

    def __init__(self,cl, name):
        PyTango.Device_4Impl.__init__(self,cl,name)
        self.debug_stream("In __init__()")
        Smaract_mcs2.init_device(self)


    def delete_device(self):
        self.debug_stream("In delete_device()")
        self.stop_polling()
        if self.socket is not None:
            del self.socket
        if self.autoproxy is not None:
            del self.autoproxy
        if self.db is not None:
            del self.db


    def init_device(self):
        self.debug_stream("In init_device()")
        self.set_state(PyTango.DevState.INIT)
        self.last_status_update = 0.0
        self.get_device_properties(self.get_device_class())
        # Sensors units are picometers/nanodegrees. We shall use milimeters/degrees.
        self.SensorUnitScaleFactor = 1e9
        self.debug_stream("Channel: " + str(self.ChannelIndex))
        self.chan = str(self.ChannelIndex)
        self.db = PyTango.Database()
        self.socket_alive = False
        try:
            # socket device connected to a SmarAct MCS2 Controller
            self.socket = PyTango.DeviceProxy(self.SocketDevice)
            self.socket.ping()
            self.socket_alive = True
            if not self.socket_check():
                self.set_state(PyTango.DevState.FAULT)
                self.status_string = "INIT failed, Socket proxy failed."
                self.debug_stream(self.status_string) 
        except PyTango.DevFailed:
            exctype, value = sys.exc_info()[:2]
            print "Failed with exception ! ", exctype
            for err in value:
                print " reason", err.reason
                print " description", err.desc
                print " origin", err.origin
                print " severity", err.severity
            self.set_state(PyTango.DevState.FAULT)
            self.status_string = err.desc
            self.debug_stream(self.status_string)
            self.socket_alive = False

        self.autoproxy = PyTango.DeviceProxy(self.get_name())
        self.attr_IsReferenced_read = False
        self.attr_IsCalibrated_read = False        
        self.attr_Position_read = 0.0
        self.attr_PositionBackup_read = 0.0
        self.attr_NextPosition_read = 0.0
        self.attr_PreviousPosition_read = 0.0
        self.attr_RequireGo_read = False
        self.attr_Acceleration_read = 0.0
        self.attr_Velocity_read = 0.0
        self.attr_HoldTime_read = 0
        self.attr_LSInvert_read = False
        self.attr_LSOffset_read = 0.0
        self.attr_HighLimitPosition_read = 0.0
        self.attr_LowLimitPosition_read = 0.0
        self.attr_StepSize_read = 0.0
        self.attr_Steps_read = 0
        self.attr_Amplitude_read = 1
        self.attr_Frequency_read = 1
        self.attr_Voltage_read = 0.0
        self.attr_SensorEnabled_read = 0
        self.attr_MotorTitle_read = ""
        self.attr_SensorStatus_read = "UNKNOWN"
        self.attr_Error_read = ""
        self.attr_MoveMode_read = None
        # if a sensor is not present it means encoderless positioner.
        self.sensor_present = False
        # Check/configure sensor type and calbirate/reference if necessary:
        self.configure_ctrl()
        if self.get_state() in (PyTango.DevState.FAULT,):
            # something went wrong...
            self.error_stream("Init failed.")
            self.status_string = "Init failed"
        else:
            self.set_state(PyTango.DevState.ON)
            self.info_stream("Device initialized.")
            self.status_string = "Device initialized"           

    # init_device() complete.

    #-----------------------------------------------------------------------------
    #    Always executed hook
    #-----------------------------------------------------------------------------
    def always_executed_hook(self):
        # self.debug_stream("In always_excuted_hook()")
        if time.time()-self.last_status_update > 0.02 and self.socket_alive:
            self.positioner_state()
            self.last_status_update = time.time()
        self.set_status(self.status_string)

    #-----------------------------------------------------------------------------
    #    Utils
    #-----------------------------------------------------------------------------

    #------------------------------------------------------------------
    #    Reconnect socket utility
    #        Returns nothing
    #------------------------------------------------------------------
    def socket_reconnect(self):
        ''' Try to reconnect the socket. ''' 
        try:
            self.socket.reconnect()
            self.debug_stream("Socket (MCS2-CTRL) reconnect called.")
        except Exception, e:
            print "Socket connection raised exception.\n"
            self.status_string = "Socket Failed: " + str(e)
            self.error_stream(self.status_string)
            self.socket_alive = False

    #------------------------------------------------------------------
    #    Checks if the controller is connected.
    #       Returns True/False
    #------------------------------------------------------------------
    def socket_check(self):
        try:
            if self.socket.state() != PyTango.DevState.ON:
                # Try to reconnect once...
                self.debug_stream("Socket not in ON, reconnect.")
                self.socket_reconnect()
                time.sleep(0.2)
                if self.socket.state() != PyTango.DevState.ON:
                    self.set_state(PyTango.DevState.FAULT)
                    self.status_string = "Socket device not in ON state. INIT failed."
                    self.debug_stream("Socket still not in ON, init failed.")
                    return False
            self.debug_stream("Socket device is in ON.")
            return True
        except:
            self.socket_alive = False
            return False 

    #------------------------------------------------------------------
    #    Write socket utility
    #        Returns nothing
    #------------------------------------------------------------------
    def socket_write(self, argin):
        # print "In ", self.get_name(), "::socket_write()"
        if not self.socket_alive:
            return
        mcs_cmd = str(argin) + END_OF_LINE
        self.debug_stream("Write cmd >> " + mcs_cmd)
        try:
            self.socket.Write(mcs_cmd)
        except Exception, e:
            print "Socket connection raised exception.\n"
            self.status_string = "Socket Failed: "+str(e)
            self.error_stream(self.status_string)
            self.attr_Error_read = self.status_string
            self.set_state(PyTango.DevState.FAULT)
            self.socket_check()

    #------------------------------------------------------------------
    #    Read/write socket utility
    #        Returns a string in case of a proper reply,
    #        0 for no reply or failure.
    #------------------------------------------------------------------
    def socket_write_and_read(self, argin):
        # print "In ", self.get_name(), "::socket_write_and_read()"
        # self.debug_stream("argin: " + str(argin))
        if not self.socket_alive:
            return        
        mcs_cmd = str(argin) + END_OF_LINE
        self.debug_stream("Write cmd >> " + mcs_cmd)
        try:
            argout = self.socket.WriteAndRead(mcs_cmd)
            if len(argout) == 0:
                self.status_string = "Cmd failed, no reply." # We ARE in the 'Sync' mode.
                self.debug_stream(self.status_string)
                return None
            argout = argout.rstrip()
            self.debug_stream("Resp << " + argout)
            return argout
        except Exception, e:
            print "Socket connection raised exception.\n"
            self.status_string = "Socket Failed: "+str(e)
            self.error_stream(self.status_string)          
            self.attr_Error_read = self.status_string
            self.set_state(PyTango.DevState.FAULT)
            self.socket_check()
            return None

    #------------------------------------------------------------------
    #    Parse MCS2 flags 
    #        argin = MCS2 state code (int)
    #        Returns a list of FLAGS that are set to True
    #------------------------------------------------------------------
    def find_active_flags(self):
        cmd_str = ":CHAN" + self.chan + ":STAT?"
        resp = self.socket_write_and_read(cmd_str)
        sensor_state = ""
        if resp is not None:
            # Find the active flags (bits)
            self.debug_stream("DevState: " + str(resp))
            int_value = int(resp)
            bits = bin(int_value)[2:][::-1]
            active_flags = []
            for i in list(range(len(bits))):
                if int(bits[i]):
                    active_flags.append(CONTROLLER_FLAGS[i])
            return active_flags
        else:
            self.status_string = "Failed to check state."
            self.error_stream(self.status_string)
            self.set_state(PyTango.DevState.ALARM)
            return None

    #------------------------------------------------------------------
    #    Get state value from the controller
    #        Parse active flags to determine the controller state
    #------------------------------------------------------------------
    def positioner_state(self):
        flags = self.find_active_flags()
        if flags is None:
            return

        self.attr_SensorStatus_read = '\n'.join(flags)

        if "SENSOR_PRESENT" not in flags:
            # Actually, this means it is an encoderless positioner,
            # Closed-loop movements not possibile
            self.sensor_present = False
            self.attr_SensorStatus_read += '\nNO_SENSOR'

        if "OVER_TEMPERATURE" in flags:
            self.set_state(PyTango.DevState.ALARM)
            self.status_string = "Overheated"
            self.error_stream(self.status_string)
            return

        if "POSITIONER_OVERLOAD" in flags:
            self.set_state(PyTango.DevState.ALARM)
            self.status_string = "Positioner overloaded"
            self.error_stream(self.status_string)
            return

        if "IS_CALIBRATED" not in flags:
            self.attr_IsCalibrated_read = False
            if "CALIBRATING" in flags:
                self.set_state(PyTango.DevState.MOVING)
                self.status_string = "Calibrating" 
                return
            else:
                self.set_state(PyTango.DevState.ALARM)                
                self.status_string = "Sensor not calibrated"
                self.error_stream(self.status_string)
                return
        else:
            self.attr_IsCalibrated_read = True

        if "IS_REFERENCED" not in flags:
            self.attr_IsReferenced_read = False
            if self.sensor_present:
                if "REFERENCING" in flags:
                    self.set_state(PyTango.DevState.MOVING)
                    self.status_string = "Referencing" 
                    return
                else:
                    self.set_state(PyTango.DevState.ALARM)                
                    self.status_string = "Sensor not referenced"
                    self.error_stream(self.status_string)
                    return
        else:
            self.attr_IsReferenced_read = True

        if "POSITIONER_FAULT" in flags:
            self.set_state(PyTango.DevState.ALARM)
            self.status_string = "Positioner fault"
            self.error_stream(self.status_string)
            return

        if "MOVEMENT_FAILED" in flags:
            if "RANGE_LIMIT_REACHED" in flags:
                self.set_state(PyTango.DevState.ON)                
                self.status_string = "Range limit reached"
                return
            elif "END_STOP_REACHED" in flags:
                self.set_state(PyTango.DevState.ON)                
                self.status_string = "End stop reached"
                return
            elif "FOLLOWING_LIMIT_REACHED" in flags:
                self.set_state(PyTango.DevState.ON)                
                self.status_string = "Following limit reached"
                return
            else:
                self.set_state(PyTango.DevState.ON)                
                self.status_string = "Movement failed"

        if "ACTIVELY_MOVING" in flags:
            self.set_state(PyTango.DevState.MOVING)
            self.status_string = "Positioning"   
        else:
            self.set_state(PyTango.DevState.ON)
            self.status_string = "Ready"   

        
    def write_sensor_type(self, value):
        cmd_str = ":CHAN" + self.chan + ":PTYP " + str(value)
        self.socket_write(cmd_str) 

    def read_sensor_type(self):
        cmd_str = ":CHAN" + self.chan + ":PTYP?"
        return self.socket_write_and_read(cmd_str)
 
    def write_move_mode(self, value):
        cmd_str = ":CHAN" + self.chan + ":MMOD " + str(value)
        self.socket_write(cmd_str) 

    def read_move_mode(self):
        cmd_str = ":CHAN" + self.chan + ":MMOD?"
        self.attr_MoveMode_read = self.socket_write_and_read(cmd_str)

    def move_abs(self, value):
        self.write_move_mode(0)
        in_val = value * self.SensorUnitScaleFactor
        cmd_str = ":MOVE" + self.chan + " " + str(int(in_val))
        self.socket_write(cmd_str)        

    def move_rel(self, value):
        self.write_move_mode(1)
        in_val = value * self.SensorUnitScaleFactor
        cmd_str = ":MOVE" + self.chan + " " + str(int(in_val))
        self.socket_write(cmd_str) 

    def move_steps(self, value):
        self.write_move_mode(4)
        cmd_str = ":MOVE" + self.chan + " " + str(int(value))
        self.socket_write(cmd_str) 
        self.attr_Position_read += value
        
    def read_error_from_ctrl(self):
        if self.socket_write_and_read(":SYST:ERR:COUN?") not in [None, 0]:
            err_code = self.socket_write_and_read(":SYST:ERR:NEXT?")
            if err_code is not None:
                self.attr_Error_read = err_code

    def configure_ctrl(self):
        ''' Sequence of operations to configure controller at init. 
        '''
        self.debug_stream('Configuring the controller...')
        if not self.socket_check():
            return

        flags = self.find_active_flags()
        if flags is None:
            # Don't know why, but regularly fails at the first attempt!?
            time.sleep(0.2)
            flags = self.find_active_flags()
            if flags is None:
                # Still no go?
                return            
        
        if "SENSOR_PRESENT" in flags:
            self.sensor_present = True
            self.debug_stream("SensorType: " + str(self.SensorType))
            self.debug_stream("InitType: " + str(self.InitType))             
        else:
            self.sensor_present = False
            # If no sensor, there is nothing to configure.
            return

        st = self.read_sensor_type()
        if int(st) != self.SensorType:
            # SensorType configured on the controller is not the correct one (owerwrite):
            self.debug_stream("Set sensor type to " + str(self.SensorType))
            self.write_sensor_type(self.SensorType)
            time.sleep(0.1)
        else: 
            self.debug_stream("Sensor type already set to " + str(self.SensorType))

        if self.InitType == 3:
            self.debug_stream("Auto-init disabled.")
            return

        if "IS_CALIBRATED" not in flags:
            self.debug_stream("Sensor must be calibrated, calibrating...")
            self.Calibrate()
            time.sleep(0.1)
            flags = self.find_active_flags()
            while "CALIBRATING" in flags:
                time.sleep(0.1)
                flags = self.find_active_flags()
            if "IS_CALIBRATED" in flags:
                self.debug_stream("Sensor successfully calibrated.")
            else:
                self.error_stream("Calibration failed!")
        else:
            self.debug_stream("Sensor already calibrated, skip calibration.")

        if "IS_REFERENCED" not in flags:
            self.debug_stream("Sensor must be referenced, referencing...")
            self.Reference()
            time.sleep(0.1)
            flags = self.find_active_flags()
            while "REFERENCING" in flags:
                time.sleep(0.1)
                flags = self.find_active_flags()
            if "IS_REFERENCED" in flags:
                self.debug_stream("Reference mark found.")
            else:
                self.error_stream("Referencing failed!")
        else:
            self.debug_stream("Sensor already referenced, skip referencing.")
            # Try to find last known position and set is PreviousPosition:
            try:
                dev_attr_props = self.db.get_device_attribute_property(self.get_name(), "PositionBackup")
                last_known_pos = float(dev_attr_props['PositionBackup']['__value'][0])
            except:
                last_known_pos = 0.0
            self.debug_stream('PreviousPosition >>> ' +str(last_known_pos))
            self.attr_PreviousPosition_read = last_known_pos                


    def read_logical_position(self):
        """ Returns angle in deg or position in mm."""
        cmd_str = ":CHAN" + self.chan + ":POS:CURR?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read Position Failed", self.status_string, "read_logical_position()")
        else:
            try:
                pos = round(float(resp)/self.SensorUnitScaleFactor, 3)
                return pos
            except Exception, e:
                PyTango.Except.throw_exception("Read Position Failed", "Unexpected reply format", "read_logical_position()")        
             
            
    #-----------------------------------------------------------------------------
    #    Smaract_mcs2 read/write attribute methods
    #-----------------------------------------------------------------------------

    def read_Error(self, attr):
        attr.set_value(self.attr_Error_read)        

    def read_Position(self, attr):
        # self.debug_stream("In read_Position()")
        if self.sensor_present:
            self.attr_Position_read = self.read_logical_position()
            attr.set_value(self.attr_Position_read)
            if self.get_state() == PyTango.DevState.ON:
                if abs(self.attr_PositionBackup_read - self.attr_Position_read) >= 0.001:
                    self.autoproxy.PositionBackup = self.attr_Position_read
        else:
            attr.set_value(self.attr_Position_read)

    def write_Position(self, attr):
        # self.debug_stream("In write_Position()")
        data = attr.get_write_value()
        if self.sensor_present:
            if self.attr_RequireGo_read :
                self.attr_NextPosition_read = data
            else:
                self.attr_PreviousPosition_read = self.read_logical_position()
                self.move_abs(data)
        else:
            self.move_steps(data - self.attr_Position_read)

    def is_Position_allowed(self, req_type):
        return self.get_state() in (PyTango.DevState.ON, PyTango.DevState.MOVING,)


    def read_Acceleration(self, attr):
        # self.debug_stream("In read_Acceleration()")
        # A value of 0 (default) deactivates the acceleration control feature.
        cmd_str = ":CHAN" + self.chan + ":ACC?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read Acceleration Failed", self.status_string, "read_Acceleration()")
        else:
            try:
                self.attr_Acceleration_read = float(resp) / self.SensorUnitScaleFactor
            except Exception, e:
                PyTango.Except.throw_exception("Read Acceleration Failed", "Unexpected reply format", "read_Acceleration()")

        attr.set_value(self.attr_Acceleration_read)

    def write_Acceleration(self, attr):
        # self.debug_stream("In write_Acceleration()")
        # A value of 0 (default) deactivates the acceleration control feature.
        data = attr.get_write_value()
        acc = int(data * self.SensorUnitScaleFactor)
        cmd_str = ":CHAN" + self.chan + ":ACC " + str(acc)
        self.socket_write(cmd_str)

    def is_Acceleration_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or (self.sensor_present and self.get_state() not in (PyTango.DevState.MOVING, PyTango.DevState.FAULT,))

    def read_Velocity(self, attr):
        # self.debug_stream("In read_Velocity()")
        # A value of 0 (default) deactivates the speed control feature.
        cmd_str = ":CHAN" + self.chan + ":VEL?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read Velocity Failed", self.status_string, "read_Velocity()")
        else:
            try:
                self.attr_Velocity_read = float(resp) / self.SensorUnitScaleFactor
            except Exception, e:
                PyTango.Except.throw_exception("Read Velocity Failed", "Unexpected reply format", "read_Velocity()")
        attr.set_value(self.attr_Velocity_read)

    def write_Velocity(self, attr):
        # self.debug_stream("In write_Velocity()")
        # A value of 0 (default) deactivates the speed control feature.
        data = attr.get_write_value()
        speed = int(data * self.SensorUnitScaleFactor)
        cmd_str = ":CHAN" + self.chan + ":VEL " + str(speed)
        self.socket_write(cmd_str)

    def is_Velocity_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or (self.sensor_present and self.get_state() not in (PyTango.DevState.MOVING, PyTango.DevState.FAULT,))

    def read_HoldTime(self, attr):
        # self.debug_stream("In read_HoldTime()")
        cmd_str = ":CHAN" + self.chan + ":HOLD?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read HoldTime Failed", self.status_string, "read_HoldTime()")
        else:
            try:
                self.attr_HoldTime_read = int(resp)
            except Exception, e:
                PyTango.Except.throw_exception("Read HoldTime Failed", "Unexpected reply format", "read_HoldTime()")
        attr.set_value(self.attr_HoldTime_read)

    def write_HoldTime(self, attr):
        # self.debug_stream("In write_HoldTime()")
        data = attr.get_write_value()
        cmd_str = ":CHAN" + self.chan + ":HOLD " + str(data)
        self.socket_write(cmd_str)

    def is_HoldTime_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or (self.sensor_present and self.get_state() not in (PyTango.DevState.MOVING, PyTango.DevState.FAULT,))

    def read_LSInvert(self, attr):
        # self.debug_stream("In read_LSInvert()")
        # A value of 0 (default) deactivates the acceleration control feature.
        cmd_str = ":CHAN" + self.chan + ":LSC:INV?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read LSInvert Failed", self.status_string, "read_LSInvert()")
        else:
            try:
                self.attr_LSInvert_read = bool(int(resp))
            except Exception, e:
                PyTango.Except.throw_exception("Read LSInvert Failed", "Unexpected reply format", "read_LSInvert()")
        attr.set_value(self.attr_LSInvert_read)

    def write_LSInvert(self, attr):
        # self.debug_stream("In write_LSInvert()")
        # Inverts logical scale count direction: 0 no / 1 yes
        data = attr.get_write_value()
        if data:
            val = 1
        else:
            val = 0
        cmd_str = ":CHAN" + self.chan + ":LSC:INV " + str(val)
        self.socket_write(cmd_str)

    def is_LSInvert_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or (self.sensor_present and self.get_state() not in (PyTango.DevState.MOVING, PyTango.DevState.FAULT,))

    def read_LSOffset(self, attr):
        # self.debug_stream("In read_LSOffset()")
        # Smaract bug: this one fails on the controller in the short form!
        # cmd_str = ":CHAN" + self.chan + ":LSC:OFF?"
        cmd_str = ":channel" + self.chan + ":lscale:offset?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read LSOffset Failed", self.status_string, "read_LSOffset()")
        else:
            try:
                self.attr_LSOffset_read = float(resp) / self.SensorUnitScaleFactor
            except Exception, e:
                PyTango.Except.throw_exception("Read LSOffset Failed", "Unexpected reply format", "read_LSOffset()")
        attr.set_value(self.attr_LSOffset_read)

    def write_LSOffset(self, attr):
        # self.debug_stream("In write_LSOffset()")
        # Sets logical scale offset.
        data = attr.get_write_value()
        val = int(data * self.SensorUnitScaleFactor)
        # Smaract bug: this one fails on the controller in the short form!
        # cmd_str = ":CHAN" + self.chan + ":LSC:OFF " + str(val)
        cmd_str = ":channel" + self.chan + ":lscale:offset " + str(val)        
        self.socket_write(cmd_str)

    def is_LSOffset_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or (self.sensor_present and self.get_state() not in (PyTango.DevState.MOVING, PyTango.DevState.FAULT,)) 

    def read_HighLimitPosition(self, attr):
        # self.debug_stream("In read_HighLimitPosition()")
        cmd_str = ":CHAN" + self.chan + ":RLIM:MAX:CURR?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read HighLimitPosition Failed", self.status_string, "read_HighLimitPosition()")
        else:
            try:
                self.attr_HighLimitPosition_read = float(resp) / self.SensorUnitScaleFactor
            except Exception, e:
                PyTango.Except.throw_exception("Read HighLimitPosition Failed", "Unexpected reply format", "read_HighLimitPosition()")        
        attr.set_value(self.attr_HighLimitPosition_read)

    def write_HighLimitPosition(self, attr):
        # self.debug_stream("In write_HighLimitPosition()")
        data = attr.get_write_value()
        val = int(data * self.SensorUnitScaleFactor)
        cmd_str = ":CHAN" + self.chan + ":RLIM:MAX:CURR " + str(val)
        self.socket_write(cmd_str)

    def is_HighLimitPosition_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or (self.sensor_present and self.get_state() not in (PyTango.DevState.MOVING, PyTango.DevState.FAULT,))

    def read_LowLimitPosition(self, attr):
        # self.debug_stream("In read_LowLimitPosition()")
        cmd_str = ":CHAN" + self.chan + ":RLIM:MIN:CURR?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read LowLimitPosition Failed", self.status_string, "read_LowLimitPosition()")
        else:
            try:
                self.attr_LowLimitPosition_read = float(resp) / self.SensorUnitScaleFactor
            except Exception, e:
                PyTango.Except.throw_exception("Read LowLimitPosition Failed", "Unexpected reply format", "read_LowLimitPosition()") 
        attr.set_value(self.attr_LowLimitPosition_read)

    def write_LowLimitPosition(self, attr):
        # self.debug_stream("In write_LowLimitPosition()")
        data = attr.get_write_value()
        val = int(data * self.SensorUnitScaleFactor)
        cmd_str = ":CHAN" + self.chan + ":RLIM:MIN:CURR " + str(val)
        self.socket_write(cmd_str)

    def is_LowLimitPosition_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or (self.sensor_present and self.get_state() not in (PyTango.DevState.MOVING, PyTango.DevState.FAULT,))

    def read_Amplitude(self, attr):
        # self.debug_stream("In read_Amplitude()")
        cmd_str = ":CHAN" + self.chan + ":STEP:AMPL?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read Amplitude Failed", self.status_string, "read_Amplitude()")
        else:
            try:
                self.attr_Amplitude_read = int(resp)
            except Exception, e:
                PyTango.Except.throw_exception("Read Amplitude Failed", "Unexpected reply format", "read_Amplitude()")
        attr.set_value(self.attr_Amplitude_read)
        self.attr_Voltage_read = round(self.attr_Amplitude_read*100.0 / float(65535), 2)

    def write_Amplitude(self, attr):
        # self.debug_stream("In write_Amplitude()")
        data = attr.get_write_value()
        cmd_str = ":CHAN" + self.chan + ":STEP:AMPL " + str(data)
        self.socket_write(cmd_str)
        self.attr_Amplitude_read = data
        self.attr_Voltage_read = round(self.attr_Amplitude_read*100.0 / float(65535), 2)

    def is_Amplitude_allowed(self, req_type):
        return not self.sensor_present        

    def read_Voltage(self, attr):
        # self.debug_stream("In read_Voltage()")
        attr.set_value(self.attr_Voltage_read)

    def write_Voltage(self, attr):
        # self.debug_stream("In write_Voltage()")
        data = attr.get_write_value()
        self.attr_Voltage_read = data
        self.autoproxy.Amplitude = int(round(self.attr_Voltage_read/100.0 * 65536))-1     

    def is_Voltage_allowed(self, req_type):
        return not self.sensor_present 

    def read_Frequency(self, attr):
        # self.debug_stream("In read_Frequency()")
        cmd_str = ":CHAN" + self.chan + ":STEP:FREQ?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read Frequency Failed", self.status_string, "read_Frequency()")
        else:
            try:
                self.attr_Frequency_read = int(resp)
            except Exception, e:
                PyTango.Except.throw_exception("Read Frequency Failed", "Unexpected reply format", "read_Frequency()")    
        attr.set_value(self.attr_Frequency_read)

    def write_Frequency(self, attr):
        # self.debug_stream("In write_Frequency()")
        data = attr.get_write_value()
        cmd_str = ":CHAN" + self.chan + ":STEP:FREQ " + str(data)
        self.socket_write(cmd_str)        
        self.attr_Frequency_read = data

    def is_Frequency_allowed(self, req_type):
        return not self.sensor_present 

    def read_MotorTitle(self, attr):
        # self.debug_stream("In read_MotorTitle()")
        attr.set_value(self.attr_MotorTitle_read)

    def write_MotorTitle(self, attr):
        # self.debug_stream("In write_MotorTitle()")
        data = attr.get_write_value()
        self.attr_MotorTitle_read = data

    def read_StepSize(self, attr):
        # self.debug_stream("In read_StepSize()")
        attr.set_value(self.attr_StepSize_read)
        # if not self.sensor_present:
        #     self.attr_Steps_read = int(round(self.attr_StepSize_read))  

    def write_StepSize(self, attr):
        # self.debug_stream("In write_StepSize()")
        data = attr.get_write_value()
        self.attr_StepSize_read = data
        # if not self.sensor_present:
        #     self.attr_Steps_read = int(round(self.attr_StepSize_read))      

    def is_StepSize_allowed(self, req_type):
        return self.sensor_present  

    def read_Steps(self, attr):
        # self.debug_stream("In read_Steps()")
        attr.set_value(self.attr_Steps_read)

    def write_Steps(self, attr):
        # self.debug_stream("In write_Steps()")
        data = attr.get_write_value()
        self.attr_Steps_read = data
        # if not self.sensor_present:
        #     self.autoproxy.StepSize = float(self.attr_Steps_read)

    def is_Steps_allowed(self, req_type):
        return not self.sensor_present             

    def read_PreviousPosition(self, attr):
        # self.debug_stream("In read_PreviousPosition()")
        attr.set_value(self.attr_PreviousPosition_read)

    def read_NextPosition(self, attr):
        # self.debug_stream("In read_NextPosition()")
        attr.set_value(self.attr_NextPosition_read)

    def write_NextPosition(self, attr):
        # self.debug_stream("In write_NextPosition()")
        data = attr.get_write_value()
        self.attr_NextPosition_read = data        

    def read_Error(self, attr):
        # self.debug_stream("In read_Error()")
        attr.set_value(self.attr_Error_read)

    def read_SensorStatus(self, attr):
        # self.debug_stream("In read_SensorStatus()")
        attr.set_value(self.attr_SensorStatus_read)

    def read_RequireGo(self, attr):
        # self.debug_stream("In read_RequireGo()")
        attr.set_value(self.attr_RequireGo_read)

    def write_RequireGo(self, attr):
        # self.debug_stream("In write_RequireGo()")
        data = attr.get_write_value()
        self.attr_RequireGo_read = data

    def read_PositionBackup(self, attr):
        # self.debug_stream("In read_PositionBackup()")
        attr.set_value(self.attr_PositionBackup_read)

    def write_PositionBackup(self, attr):
        # self.debug_stream("In write_PositionBackup()")
        data = attr.get_write_value()
        self.attr_PositionBackup_read = data

    def is_PositionBackup_allowed(self, req_type):
        return self.sensor_present 

    def read_SensorPowerMode(self, attr):
        # self.debug_stream("In read_SensorEnabled()")
        cmd_str = ":CHAN" + self.chan + ":SENS:MODE?"
        resp = self.socket_write_and_read(cmd_str)
        if resp is None:
            # self.read_error_from_ctrl()
            PyTango.Except.throw_exception("Read Sensor Enabled Failed", self.status_string, "read_SensorEnabled()")
        else:
            try:
                self.attr_SensorEnabled_read = int(resp)
            except Exception, e:
                PyTango.Except.throw_exception("Read Sensor Enabled Failed", "Unexpected reply format", "read_SensorEnabled()")
        attr.set_value(self.attr_SensorEnabled_read)

    def write_SensorPowerMode(self, attr):
        # self.debug_stream("In write_SensorEnabled()")
        data = attr.get_write_value()
        cmd_str = ":CHAN" + self.chan + ":SENS:MODE " + str(data)
        self.socket_write(cmd_str)

    def is_SensorPowerMode_allowed(self, req_type):
        return req_type is PyTango.AttReqType.READ_REQ or (self.sensor_present and self.get_state() not in (PyTango.DevState.MOVING, PyTango.DevState.FAULT,))

    def read_IsReferenced(self, attr):
        attr.set_value(self.attr_IsReferenced_read)

    def read_IsCalibrated(self, attr):
        attr.set_value(self.attr_IsCalibrated_read)


    #-----------------------------------------------------------------------------
    #    Smaract_mcs2 command methods
    #-----------------------------------------------------------------------------

    def On(self):
        #self.debug_stream("In On()")
        if self.get_state() in (PyTango.DevState.OFF,):
            self.set_state(PyTango.DevState.ON)
        else:
            pass

    def Off(self):
        # self.debug_stream("In Off()")
        self.set_state(PyTango.DevState.OFF)


    def Abort(self):
        self.debug_stream("In Abort()")
        # Same as STOP
        cmd_str = ":STOP" + self.chan
        self.socket_write(cmd_str)
        

    def Calibrate(self):
        self.debug_stream("In Calibrate()")
        # if self.attr_IsCalibrated_read:
        #     self.debug_stream("Already calibrated. Do nothing.")
        #     return      
        cmd_str = ":CHAN" + self.chan + ":CAL:OPT 0"
        self.socket_write(cmd_str)
        # start actual calibration:
        cmd_str = ":CAL" + self.chan
        self.socket_write(cmd_str)


    def Reference(self):
        self.debug_stream("In Reference()")
        # if self.attr_IsReferenced_read:
        #     self.debug_stream("Already referenced. Do nothing.")
        #     return
        # Evaluate the last known physical position based on the values stored in tango db 
        # and pick the referencing direction accordingly:
        # Referencing/Calibrating mode 0 is forward (default), 1 is backward
        ref_mode = 0

        try:
            dev_attr_props = self.db.get_device_attribute_property(self.get_name(), "PositionBackup")
            last_known_pos = float(dev_attr_props['PositionBackup']['__value'][0])
        except:
            last_known_pos = 0.0
        self.debug_stream('Last Known Position >>> ' +str(last_known_pos))
        self.attr_PreviousPosition_read = last_known_pos
        try:
            dev_attr_props = self.db.get_device_attribute_property(self.get_name(), "LSOffset")
            ls_offset = float(dev_attr_props['LSOffset']['__value'][0])
        except:
            ls_offset = 0.0
        self.debug_stream('Logical scale offset >>> ' +str(ls_offset))
        try:
            dev_attr_props = self.db.get_device_attribute_property(self.get_name(), "LSInvert")
            ls_invert = bool(strtobool(dev_attr_props['LSInvert']['__value'][0]))
        except:
            ls_invert = False
        self.debug_stream('Logical scale inversion >>> ' +str(ls_invert))
        if ls_invert:
            scale_dir = -1.0
        else:
            scale_dir = 1
            
        if self.InitType == 0:
            ref_mode = 0
        elif self.InitType == 1:
            ref_mode = 1
        elif self.InitType in [2, 3]:
            if ls_offset + scale_dir * last_known_pos <= 0:
                ref_mode = 0
            else:
                ref_mode = 1
        else:
            return

        # set reference options (0 search fwd / 1 backward)
        cmd_str = ":CHAN" + self.chan + ":REF:OPT " + str(ref_mode)
        self.socket_write(cmd_str)
        # start actual referencing:
        cmd_str = ":REF" + self.chan
        self.socket_write(cmd_str)

    def is_Reference_allowed(self):
        return self.sensor_present 


    def ReferenceForward(self):
        self.debug_stream("In ReferenceForward()")
        # if self.attr_IsReferenced_read:
        #     self.debug_stream("Already referenced. Do nothing.")
        #     return
        ref_mode = 0
        cmd_str = ":CHAN" + self.chan + ":REF:OPT " + str(ref_mode)
        self.socket_write(cmd_str)
        # start actual referencing:
        cmd_str = ":REF" + self.chan
        self.socket_write(cmd_str)

    def is_ReferenceForward_allowed(self):
        return self.sensor_present        


    def ReferenceBackward(self):
        self.debug_stream("In ReferenceBackward()")
        # if self.attr_IsReferenced_read:
        #     self.debug_stream("Already referenced. Do nothing.")
        #     return
        ref_mode = 1
        cmd_str = ":CHAN" + self.chan + ":REF:OPT " + str(ref_mode)
        self.socket_write(cmd_str)
        # start actual referencing:
        cmd_str = ":REF" + self.chan
        self.socket_write(cmd_str)

    def is_ReferenceBackward_allowed(self):
        return self.sensor_present 


    def Go(self):
        # self.debug_stream("In Go()")
        if self.attr_NextPosition_read != self.attr_Position_read:
            self.attr_PreviousPosition_read = self.read_logical_position()
            self.move_abs(self.attr_NextPosition_read)

    def is_Go_allowed(self):
        return self.sensor_present and self.get_state() in (PyTango.DevState.ON,)    


    def StepUp(self):
        # self.debug_stream("In StepUp()")
        if self.sensor_present:
            if self.attr_StepSize_read == 0:
                self.status_string = 'Step size is not set.'
                self.set_status(self.status_string)
                return               
            if self.attr_RequireGo_read :
                curr_pos = self.autoproxy.Position
                self.attr_NextPosition_read = curr_pos + self.attr_StepSize_read
                return
            self.attr_PreviousPosition_read = self.read_logical_position()
            self.move_rel(self.attr_StepSize_read)
        else: 
            if self.attr_Steps_read == 0:
                self.status_string = 'Steps are not set.'
                self.set_status(self.status_string)
                return                
            self.move_steps(self.attr_Steps_read)

    def is_StepUp_allowed(self):
        return self.get_state() in (PyTango.DevState.ON,)     


    def StepDown(self):
        # self.debug_stream("In StepDown()") 
        if self.sensor_present:
            if self.attr_StepSize_read == 0:
                self.status_string = 'Step size is not set.'
                self.set_status(self.status_string)
                return       
            if self.attr_RequireGo_read :
                curr_pos = self.autoproxy.Position
                self.attr_NextPosition_read = curr_pos - self.attr_StepSize_read
                return
            self.attr_PreviousPosition_read = self.read_logical_position()
            self.move_rel(-1 * self.attr_StepSize_read)
        else:   
            if self.attr_Steps_read == 0:
                self.status_string = 'Steps are not set.'
                self.set_status(self.status_string)
                return     
            self.move_steps(-1 * self.attr_Steps_read)

    def is_StepDown_allowed(self):
        return self.get_state() in (PyTango.DevState.ON,)         


    # def AdminState(self, argin):
    #     """
    #     :param argin:
    #     :type: PyTango.DevString
    #     :return:
    #     :rtype: PyTango.DevVoid """
    #     # self.debug_stream("In AdminState()")
    #     if argin == 'kik':
    #         self.debug_stream('Admin rights granted.')
    #     else:
    #         self.warn_stream('Admin rights refused, incorrect password.')
    #         self.status_string = "Incorrect password"
    #         PyTango.Except.throw_exception("Admin access denied.", self.status_string, "AdminState()")


    def DefinePosition(self, def_pos):
        """ Defines the scale shift. Sets the current physial position
            to the logical value provided in input. 
        """
        # self.debug_stream("In DefinePosition()")
        if def_pos == self.attr_Position_read:
            return
        # logical_pos = self.attr_Position_read 
        # scale_shift = self.attr_LSOffset_read
        # Must take into account the iverted axis direction
        if self.attr_LSInvert_read:
            physical_pos = self.attr_LSOffset_read - self.attr_Position_read
            new_shift = physical_pos + def_pos
        else:    
            physical_pos = self.attr_LSOffset_read + self.attr_Position_read
            new_shift = physical_pos - def_pos
        val = int(new_shift * self.SensorUnitScaleFactor)
        cmd_str = ":channel" + self.chan + ":lscale:offset " + str(val)
        self.socket_write(cmd_str)        
        
    def is_DefinePosition_allowed(self):
        return self.sensor_present and self.get_state() in (PyTango.DevState.ON,)   


    def GoToHighLimit(self):
        # self.debug_stream("In GoToHighLimit()")
        if self.attr_HighLimitPosition_read == self.attr_Position_read or \
            self.attr_HighLimitPosition_read == 0.0:
            return
        self.move_abs(self.attr_HighLimitPosition_read)

    def is_GoToHighLimit_allowed(self):
        return self.sensor_present and self.get_state() in (PyTango.DevState.ON,)        


    def GoToLowLimit(self):
        # self.debug_stream("In GoToLowLimit()")
        if self.attr_LowLimitPosition_read == self.attr_Position_read or \
            self.attr_LowLimitPosition_read == 0.0:
            return
        self.move_abs(self.attr_LowLimitPosition_read)

    def is_GoToLowLimit_allowed(self):
        return self.sensor_present and self.get_state() in (PyTango.DevState.ON,)


    def Stop(self):
        # self.debug_stream("In Stop()")
        cmd_str = ":STOP" + self.chan
        self.socket_write(cmd_str)


    def Reset(self):
        # self.debug_stream("In Reset()")
        self.debug_stream("In Reset() == Init()")
        self.autoproxy.Init()


    def write_and_read_command(self, argin):
        """ This command will be used to iteract with the motor in a console like maner.
        It must return either an empty string, or a string with the answer.
        It is advisable not to add a new line at the end of the answer.

        :param argin:
        :type: PyTango.DevString
        :return:
        :rtype: PyTango.DevString """
        # self.debug_stream("In write_and_read_command()")
        argout = ''
        if '?' in argin:
            argout = self.socket_write_and_read(argin)
            if argout == 0:
                # self.read_error_from_ctrl()
                PyTango.Except.throw_exception("Write_and_read Cmd Failed", self.status_string, "write_and_read_command()")
        else:
            self.socket_write(argin)
            argout = ''
        return argout


class Smaract_mcs2Class(PyTango.DeviceClass):
    #--------- Add you global class variables here --------------------------

    def dyn_attr(self, dev_list):
        """Invoked to create dynamic attributes for the given devices.
        Default implementation calls
        :meth:`Smaract_mcs2.initialize_dynamic_attributes` for each device

        :param dev_list: list of devices
        :type dev_list: :class:`PyTango.DeviceImpl`"""

        # Have none
        return

        for dev in dev_list:
            try:
                dev.initialize_dynamic_attributes()
            except:
                import traceback
                dev.warn_stream("Failed to initialize dynamic attributes")
                dev.debug_stream("Details: " + traceback.format_exc())


    #    Class Properties
    class_property_list = {
        }


    #    Device Properties
    device_property_list = {
        'SocketDevice':
            [PyTango.DevString,
            "Tango Socket Device that connects to a physical device.",
            []],
        'ChannelIndex':
            [PyTango.DevShort,
            "Integer number of MCS2 controller channel (0-2/0-5/0-11)",
            []],
        'SensorType':
            [PyTango.DevShort,
            "Type of sensor attached to the MCS controller channel.",
            []],
        'InitType':
            [PyTango.DevShort,
            "0: forward referencing, 1: backward referencing, 2: auto calibrating and referencing, 3: No auto-init.",
            []],
        }


    #    Command definitions
    cmd_list = {
        'On':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Off':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Abort':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Calibrate':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Reference':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'ReferenceForward':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'ReferenceBackward':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Go':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'StepUp':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'StepDown':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        # 'AdminState':
        #     [[PyTango.DevString, "none"],
        #     [PyTango.DevVoid, "none"]],
        'DefinePosition':
            [[PyTango.DevDouble, "none"],
            [PyTango.DevVoid, "none"]],
        'GoToHighLimit':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'GoToLowLimit':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Stop':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'Reset':
            [[PyTango.DevVoid, "none"],
            [PyTango.DevVoid, "none"]],
        'write_and_read_command':
            [[PyTango.DevString, "none"],
            [PyTango.DevString, "none"]],
        }


    #    Attribute definitions
    attr_list = {
        'Position':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Logical position in mm for linear sensors and \
                in degrees for rotary sensors. Depends on physical position, \
                scale shift and inverted axis direction. For sensorless positioners \
                it is simply a steps counter."
            }],
        'Acceleration':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'max value': "10.0",
                'min value': "0.0",
                'description': "Defines the acceleration in mm/s^2 for linear \
                positioners and in degrees/s^2 for rotary positioners. \
                The valid range is 0..10. A value of 0 (default) deactivates \
                the acceleration control feature.",
                'memorized':"true",
            }],
        'Velocity':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'max value': "100.0",
                'min value': "0.0",
                'description': "Defines the speed in mm/s for linear positioners \
                and in degrees/s for rotary positioners. The valid range is 0..100. \
                A value of 0 (default) deactivates the speed control feature.",
                'memorized':"true",
            }],
        'Frequency':
            [[PyTango.DevLong,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'max value': "20000",
                'min value': "1",
                'description': "Open-loop steps frequency in Hertz.\
                The default value is 100. Used in stepper mode only.",
                'memorized':"true",
            }],
        'Amplitude':
            [[PyTango.DevLong,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'max value': "65535",
                'min value': "1",
                'description': "Voltage in binary form, the valid range is 1..65535 \
                A value of 65535 (max) equals to 100 Volts. Lower amplitude values \
                result in a smaller step width. Used in stepper mode only.",
                'memorized':"true",
            }],
        'Voltage':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'max value': "100",
                'min value': "1.6e-5",
                'description': "Voltage in analog form derived from Amplitude. \
                Used in stepper mode only.",
            }],
        'Steps':
            [[PyTango.DevLong,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Number of steps \
                in the stepper mode (no sensor).",
                'memorized':"true",
            }],
        'StepSize':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Step size in mm for linear sensors, \
                in degrees for rotary sensors.",
                'memorized':"true",
            }],
        'HoldTime':
            [[PyTango.DevShort,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'max value': "5000",
                'min value': "0",
                'description': "This property specifies how long (in ms) the position \
                is actively held after reaching the target position. \
                After the hold time elapsed the channel is stopped and \
                the control-loop is disabled.",
                'memorized':"true",
            }],
        'MotorTitle':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Motor's given name for the GUI's title label.",
                'memorized':"true",
            }],
        'LSInvert':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "If set, logical scale count direction is inverted.",
                'memorized':"false",
            }],
        'LSOffset':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Logical scale offset in mm/degrees.",
                'memorized':"false",
            }],
        'SensorStatus':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description': "String representation of active flags.",
            }],
        'NextPosition':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Next position in mm for linear sensors and \
                in degrees for rotary sensors. Requires go() command to move.",
            }],
        'PreviousPosition':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description': "Previous position held, in mm for \
                linear sensors and in degrees for rotary sensors.",
            }],
        'Error':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description': "Description of the last error occured.",
            }],
        'RequireGo':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "If selected, write Position is disabled. \
                Instead, the next position attribute must be set and \
                Go command executed.",
                'memorized':"true",
            }],
        'HighLimitPosition':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "High limit position in mm for linear sensors \
                and in degrees for rotary sensors.",
                'memorized':"true",
            }],
        'LowLimitPosition':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Low limit position in mm for linear sensors \
                and in degrees for rotary sensors.",
                'memorized':"true",
            }],
        'PositionBackup':
            [[PyTango.DevDouble,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description': "Last known position in mm for linear sensors \
                and in degrees for rotary sensors.",
                'memorized':"true"
            }],
        'SensorPowerMode':
            [[PyTango.DevShort,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'max value':"2",
                'min value':"0",
                'description': "Selects the mode. Must be either \
                0 (disabled), 1 (enabled) or 2 (powersave).",
                'memorized':"true",
            }],
        'IsCalibrated':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description': "Checks if the sensor has been calibrated. \
                If not, calibration is necessary.",
            }],
        'IsReferenced':
            [[PyTango.DevBoolean,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description': "Checks if the sensor knows its physical \
                position. If not, reference mark must be searched for.",
            }],
        }


def main():
    try:
        py = PyTango.Util(sys.argv)
        py.add_class(Smaract_mcs2Class,Smaract_mcs2,'Smaract_mcs2')

        U = PyTango.Util.instance()
        U.server_init()
        U.server_run()

    except PyTango.DevFailed,e:
        print '-------> Received a DevFailed exception:',e
    except Exception,e:
        print '-------> An unforeseen exception occured....',e

if __name__ == '__main__':
    main()
